/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version
 * 2.1 of the License.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * END COPYRIGHT BLOCK **/
#ifndef SETUPLDAP_H
#define SETUPLDAP_H

#ifdef __cplusplus
extern "C" {
#endif
/*
****************************************************************
** Ldap Object
*****************************************************************
* Create an Ldap Object. binddn and bindpwd are optional so specify
* NULL if they are not present
*/

NSAPI_EXTERN LdapErrorCode setupCreateLdap(Ldap ** ldap,
                                   const char *ldapURL,
                                   const char *ldapUser,
                                   const char *ldapPwd,
                                   const char *binddn,
                                   const char *bindpwd);
NSAPI_EXTERN void          setupDestroyLdap (Ldap *);
/*
 * Retrieve the "true" Bind DN of the LDAP user
 */
NSAPI_EXTERN const char    *setupGetUserDN (const Ldap *);
/*
 * Retrieve the LDAP server version
 */
NSAPI_EXTERN const char    *setupGetServerVersion(const Ldap *);
/*
****************************************************************
** Ldap Entry object
*****************************************************************
*/
NSAPI_EXTERN LdapEntry     *setupCreateLdapEntry(const Ldap *);
NSAPI_EXTERN void          setupDestroyLdapEntry(LdapEntry *);
/*
 * Retrieve the DN associated with the entry.
 */
NSAPI_EXTERN const char    *setupGetEntryDN (const LdapEntry *ldapEntry);
/*
 * Determine if the entry as specified by the entryDN exists (at least one)
 */
NSAPI_EXTERN Bool          setupEntryExists(LdapEntry *ldapEntry, const char *entryDN);
NSAPI_EXTERN LdapErrorCode setupRetrieveLdapEntry(LdapEntry *ldapEntry, const char *entryDN);
NSAPI_EXTERN LdapErrorCode setupRetrieveLdapEntries(LdapEntry *ldapEntry,
                           const char *criteria, int scope, const char *baseDN);
NSAPI_EXTERN LdapErrorCode setupNextLdapEntry(LdapEntry *ldapEntry);
/*
 * Get all the attribute names in this entry
 */
NSAPI_EXTERN char **setupGetLdapAttributeNames(const LdapEntry *);
NSAPI_EXTERN void setupFreeLdapAttributeNames(LdapEntry *, char **attrNames);

/*
 * Giving the name of an attribute, return the NULL terminated list of
 * associated attribute value(s). Just use freeAttributes() to free the
 * returned list after use.
 */
NSAPI_EXTERN char **setupGetLdapAttributes(const LdapEntry *, const char *name);
NSAPI_EXTERN void setupFreeLdapAttributes(LdapEntry *, char **attributes);
NSAPI_EXTERN const char       *setupGetEntryDN(const LdapEntry *);

/*
 * Setting attributes. For single-value attribute, use setAttribute. For
 * attributes that are typically multivalue (as you know), especially
 * objectclass, use setAttributes(). Don't mix setAttribute and setAttributes
 * on the same attribute - I am not that smart.
 */
NSAPI_EXTERN void setupSetLdapAttribute(LdapEntry *, const char *name, const char *value);
NSAPI_EXTERN void setupAddLdapAttribute(LdapEntry *, const char *name, const char *value);
NSAPI_EXTERN void setupAddLdapAttributes(LdapEntry *, const char *name,const char **values);

NSAPI_EXTERN LdapErrorCode setupRenameLdapEntry(LdapEntry *, const char *entryDN, const char *rdn);

NSAPI_EXTERN LdapErrorCode setupInsertLdapEntry(LdapEntry *, const char *entryDN);

NSAPI_EXTERN LdapErrorCode setupUpdateLdapEntry(LdapEntry *, const char *entryDN);

NSAPI_EXTERN LdapErrorCode setupReplaceLdapEntry(LdapEntry *, const char *entryDN);

/*
 * Drop an entry
 */
NSAPI_EXTERN LdapErrorCode setupDropLdapEntry(Ldap *, const char *entryDN);
NSAPI_EXTERN LdapErrorCode setupDropAllLdapEntries(Ldap *, const char *entryDN);

/*
****************************************************************
** SIE/ISIE creation routines
*****************************************************************
*/

/*
 * setupInstallFramework()
 */

NSAPI_EXTERN LdapErrorCode setupInstallFramework(const Ldap *, const char *adminDomain);

/*
 * createSIE()
 *    Register a Server Instance Entry specified by sieEntry,
 *    associated with the NetscapeApplication specified by isieDN.
 *    The entry is to be inserted at subTreeDN.
 *
 *    This routine will perform the following:
 *       - Create an installed software instance under NetscapeRoot if
 *         it has not been created, otherwise, update the entry with
 *         appropriate information
 *
 *       - Generate the appropriate SIE entry's common
 *         name to guarantee the uniqueness of the SIE. A seeAlso
 *         link will be created from the SIE to the ISIE and vice versa.
 *
 *   Parameters:
 *     - sieEntry: LDAP entry for the SIE
 *     - appEntry: LDAP entry for the ISIE
 *     - hostName: name of this machine, does not need to be FQDN, BUT
 *        appEntry's serverHostName attribute must contain the FQDN
 *     - installedLocation: the location where this product is installed
 *     - subTreeDN: the DN of the subtree where the SIE is to be installed.
 *          The default is ou=Netscape Suitespot,<base-suffix>
 *   Absolute required attributes are:
 *    * appEntry
 *        - objectclass = netscapeApplication
 *        - name = product name, e.g. Netscape Enterprise Server
 *        - version = product version , e.g. 4.0a0
 *        - other attributes from the Product Information File.
 *    * sieEntry
 *        - objectclass (netscapeXXXServer, e.g. netscapeMailServer
 *               which must be inherited from NetscapeServer
 *        - sie's password
 *        - serverID
 *
 * Hint:
 *   use entryDN() to get the full DN of the ISIE and SIE
 *   and save them for later use.
 */
NSAPI_EXTERN LdapErrorCode setupCreateSIE(
         LdapEntry *sieEntry,
         LdapEntry *appEntry,
	 const char *hostName,
         const char *installedLocation,
         const char *adminDomain);

/*
 * removeSIE
 *   Remove the server instance, and the pointer to it from ISIE.
 *   Optionally, removeISIE can be set to True to remove the ISIE also.
 *   In general, removeISIE flag should always set to False. Only set it
 *   to True when used in binaries uninstallation code
 */

NSAPI_EXTERN LdapErrorCode setupRemoveSIE(Ldap *, const char *sieDN, Bool removeISIE);

/*
 * removeISIE
 *   Remove the installed software instance. To be used by uninstallation
 *   code.
 *   The ISIE can only be removed when there are no associated server
 *   instances.  The routine will remove the whole SIE hiearchy if
 *   possible up to bot not including netscapRoot.
 */
NSAPI_EXTERN LdapErrorCode setupRemoveISIE(Ldap *, const char *isieDN);

/*
 * updateISIEVersionInfo
 *   Update the ISIE with the version information as well as other related
 *   information. Used by the setup program only
 */
NSAPI_EXTERN LdapErrorCode setupUpdateISIEVersionInfo (
         Ldap *,
         LdapEntry *,
         const char *productName,
         const char *hostName,
         const char *installedLocation,
         const char *adminDomainName);

/*
 * createACIForConfigData
 *   Set the appropriate access control for the LDAP entry that contains
 *   modifiable configuration data. e.g. cn=Configuration, cn=sie-cn, ou=
 *   Netscape Suitespot, o=example.com
 *     sieDN must be a valid DN for an existing SIE entry.
 *
 */
NSAPI_EXTERN LdapErrorCode setupCreateACIForConfigEntry (LdapEntry *configEntry, const char *sieDN);

/*
****************************************************************
** Miscellaneous routines
*****************************************************************
*/

/*
 * printEntry: print entry to fileName. set fileName to NULL
 *     to print to stdout
 */
NSAPI_EXTERN void setupPrintLdapEntry(const LdapEntry *, const char *fileName);

/*
 * verifyLdapUrl
 *   verify the validity of an LDAP URL (format, ldap server, etc...)
 */
NSAPI_EXTERN  LdapErrorCode setupVerifyLdapUrl(
         const char *url,
         const char *bindDN,
         const char *bindPwd);

/*
 * authLdapUser()
 *   Just authenticate an Ldap user. This is just like a wrapper
 *   for createLdap() without returning the Ldap object.
 */
NSAPI_EXTERN  LdapErrorCode setupAuthLdapUser(
         const char *url,
         const char *ldapUser,
         const char *ldapPwd,
         const char *bindDN,
         const char *bindPwd);

/*
 * isSubTreeWritable
 *  Test whether a subtree is writable
 */
NSAPI_EXTERN LdapErrorCode setupIsSubTreeWritable(const Ldap *, const char *subTreeDN);

/*
 * setupUpdateLdifEntries
 * setupReplaceLdifEntries
 * setupAddLdifEntries
 *
 *   Update/Replace/Add Ldap entries specified in an LDIF-like file into LDAP
 *   below a given DN. In all these cases, if the entries do not exist, they
 *   will be added; otherwise, the entries will be updated for
 *   setupUpdateLdifEntries or replaced with setupReplaceLdifEntries
 *
 *   Ldap - pointer to a valid Ldap struct
 *   subTreeDN - the DN of the location to which the entries will be
 *     inserted. If the entry in the LDIF file has a DN attribute,
 *     this DN attribute should be a relative DN and will be
 *     prepended to subTreeDN to form the location where the entry is
 *     to be inserted. If subTreeDN is NULL, the DN attribute found
 *     in the entry will be used as an absolute DN to insert the
 *     entry. If subTreeDN is not NULL and there's no dn attribute
 *     in the LDIF entry, subTreeDN will be used as an absolute DN to
 *     insert the entry to. It's an error if both subTreeDN and the
 *     DN attribute are missing, in which case the entry is simply
 *     skipped.
 *
 *   ldifFile  - name of the LDIF file
 *   adminSIEDN - NULL in general. If not NULL, specify the DN of the
 *      Admin's SIE to be appended to the values of all className
 *      attributes found in the  Ldif entries.
 *
 */
NSAPI_EXTERN LdapErrorCode setupUpdateLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN);

NSAPI_EXTERN LdapErrorCode setupReplaceLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN);

NSAPI_EXTERN LdapErrorCode setupAddLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN);

NSAPI_EXTERN LdapErrorCode setupInsertSchemaEntries (Ldap *ldap, const char *schemaFile);

/*
 * guessAdminDomain
 *   Guess the Administration Domain that might exist on this directory server
 */
NSAPI_EXTERN char * setupGuessAdminDomainName(const char *ldapURL, const char *ldapUser, const char *ldapPwd);
/*
 * checkAdminDomain
 *   check the validity of the Administration Domain that exists on this directory server
 */
NSAPI_EXTERN LdapErrorCode setupCheckAdminDomain(const Ldap *ldap, const char *adminDomain);

/*
 * addUserDirectory
 *   Configure a domain with a User Directory
 */
NSAPI_EXTERN LdapErrorCode setupAddGlobalUserDirectory (const Ldap *ldap, const char *domain, const char *userDirUrl, const char *bindDN, const char *bindPwd);

/*
 * getUserDirectory
 *   Get the user Directory associated with this SIE
 */
NSAPI_EXTERN LdapErrorCode setupGetUserDirectory(const Ldap *ldap, const char *sieDN, char **url, char **bindDN, char **bindPwd);

/*
 * getLocalUserDirectory
 *   Get the User Directory associated with this server root
 */
NSAPI_EXTERN LdapErrorCode setupGetLocalUserDirectory (const Ldap *,
   const char *domain, const char *hostName, const char *serverRoot, char **userDirUrl, char **bindDN, char **bindPwd);

/*
 * getGlobalUserDirectory
 *   Get the User Directory associated with this domain
 */
NSAPI_EXTERN LdapErrorCode setupGetGlobalUserDirectory (const Ldap *,
   const char *domain, char **userDirUrl, char **bindDN, char **bindPwd);

NSAPI_EXTERN LdapErrorCode setupCreateSuffix(Ldap *ldap, char *suffix);
/* Use setupCreateSuffixAndBackend for DS5.0 and higher */
NSAPI_EXTERN LdapErrorCode setupCreateSuffixAndBackend(Ldap *ldap, char *suffix, char *name);
/* DS50 and higher requires a unique name for each suffix. Use
 * setupGenBackend with a prefix and it'll give you a unique name. Eg:
 * give it: MyServer and it might return something like: MyServer15
 * which you can then pass to setupCreateSuffixAndBackend*/
NSAPI_EXTERN char * setupGenBackendName(Ldap *ldap, char *suffix, char *prefix);

NSAPI_EXTERN LdapErrorCode setupLdapModify(Ldap *ldap, const char *dn, LDAPMod **mods);

#ifdef __cplusplus
}
#endif

#endif
