/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version
 * 2.1 of the License.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * END COPYRIGHT BLOCK **/
/*********************************************************************
**
** NAME:
**   schema.cpp
**
** DESCRIPTION:
**   Import Schema configuration file used during installation.
**
** NOTES:
**
*********************************************************************/

extern "C" {
#if defined(AIX)
   #include <strings.h>
#endif
#include <string.h>
#include <ctype.h>
};

#include "ldapu.h"
#include "setupapi.h"


#define CONFIG_ATTRIBUTETYPES "attribute"
#define SCHEMA_ATTRIBUTETYPES "attributetypes"

#define CONFIG_OBJECTCLASSES "objectclass"
#define SCHEMA_OBJECTCLASSES "objectclasses"

#define SCHEMA_DN  "cn=schema"


struct SYNTAXTOOID {
   char *syntax;
   char *oid;
} SyntaxToOID[] = {
   {(char *)"cis", (char *)"1.3.6.1.4.1.1466.115.121.1.15" },
   {(char *)"telephone", (char *)"1.3.6.1.4.1.1466.11.121.1.50" },
   {(char *)"tel", (char *)"1.3.6.1.4.1.1466.11.121.1.50" },
   {(char *)"dn", (char *)"1.3.6.1.4.1.1466.115.121.1.12" },
   {(char *)"ces", (char *)"1.3.6.1.4.1.1466.115.121.1.26" },
   {(char *)"integer", (char *)"1.3.6.1.4.1.1466.115.121.1.27" },
   {(char *)"int", (char *)"1.3.6.1.4.1.1466.115.121.1.27" },
   {(char *)"binary", (char *)"1.3.6.1.4.1.1466.115.121.1.5" },
   {(char *)"bin", (char *)"1.3.6.1.4.1.1466.115.121.1.5" },
   {NULL, NULL},
};

const char * LdapSchema::syntaxToOID(const char *syntax)
{
   struct SYNTAXTOOID *p;

   p = SyntaxToOID;

   while (p->syntax && strcasecmp(syntax, p->syntax))
   {
      p ++ ;
   }

   if (p->syntax)
      return p->oid;
   else
      return NULL;
}


LdapSchema::LdapSchema(Ldap *ld)
{
   _at.clear();
   _oc.clear();
   _schemaFile = NULL;
   _ldap = ld;
   _ldapEntry = NULL;
}

LdapSchema::~LdapSchema()
{
   _at.clear();
   _oc.clear();
   _schemaFile = NULL;

   if (_ldapEntry != NULL)
      delete _ldapEntry;
}

char *LdapSchema::getSchemaName(char *value)
{
   char *name;
   char *p, *q;
   int len;

   p = strstr(value, " NAME ");
   if (p == NULL)
      return NULL;

   p = p + strlen(" NAME ");
   q = strchr (p, ' ');
   if (q == NULL)
      return NULL;


   len = q - p + 1;
   name = (char *) malloc(len);
   memset(name, 0, len);

   // strip single quote
   if (*p == '\'')
      p ++;
   if (*(q-1) == '\'')
      q --;
   strncpy(name, p, q - p);

   return name;
}


int LdapSchema::retrieve()
{
   LdapError err;


   _ldapEntry = new LdapEntry(_ldap);

   if ((err = _ldapEntry->retrieve("cn=schema")) == OKAY)
   {
      char **attr;
      char **value;
      int i;
      char *name;

      attr = _ldapEntry->getAttributeNames();

      value = _ldapEntry->getAttributes("attributetypes");
      for (i = 0; value[i] != NULL; i ++)
      {
         name = getSchemaName(value[i]);
         if (name)
         {
            _at.set(name, value[i]);
            free(name);
         }
      }

      value = _ldapEntry->getAttributes("objectclasses");
      for (i = 0; value[i] != NULL; i ++)
      {
         name = getSchemaName(value[i]);
         if (name)
         {
            _oc.set(name, value[i]);
            free(name);
         }
      }
   }

   return err;
}

const char * LdapSchema::getValue(char *type, char *name)
{
   if (strcasecmp(type, SCHEMA_ATTRIBUTETYPES) == 0)
      return _at.get(name);
   else if (strcasecmp(type, SCHEMA_OBJECTCLASSES) == 0)
      return _oc.get(name);

   return NULL;
}


int LdapSchema::domodify(LDAPMod **mods)
{
   return _ldap->modify(SCHEMA_DN, mods);
}

/* 
 * isBlank and stripBlank are from ldapu.cpp
 */
Bool
LdapSchema::isBlank(const char *buf)
{
   Bool ret = True;

   while (*buf != 0)
   {
      char ch;
      ch = *buf++;
      if (!i18n_isspace(ch))
      {
         ret = False;
         break;
      }
   }
   return ret;
}

void 
LdapSchema::stripBlank(char *buf, Bool bothEnds)
{
   char *s = buf;
   int newlen = strlen(buf);

   while (i18n_isspace(*s))
   {
      ++s;
   }

   if (s > buf)
   {
      newlen -= (s - buf);
      memcpy(buf, s, newlen);
      buf[newlen] = '\0';
   }

   s = buf + newlen - 1;
   if (bothEnds)
   {
      while (i18n_isspace(*s))
      {
         --s;
      }
      *(s + 1) = '\0';
   }
}

#define ADDFORMVALUE(entryValue, curEntryValueP, curValLen, form, value) \
{ \
  int len; \
  len = snprintf((curEntryValueP), (curValLen), (form), (value)); \
  if (!memchr((curEntryValueP), 0, (curValLen))) return False; /*overflow*/ \
  len = strlen(entryValue); \
  (curValLen) = (entryValueLen) - len; \
  (curEntryValueP) = (entryValue) + len; \
}

Bool
LdapSchema::formValue(char *type, NVPair *nv,
                      char *entryValue, int entryValueLen)
{
   const char *s;
   char *curEntryValueP = entryValue;
   int curValLen = entryValueLen;
   int curLen;

   if (strcasecmp(type, SCHEMA_ATTRIBUTETYPES) == 0)
   {
      /* 
       * should get oid=...
       *            name=...
       *            desc=...
       *            syntax=...
       *            alias=...
       *
       * syntax:
       *    ( %s NAME '%s' DESC '%s' SYNTAX '%s' SINGLE-VALUE )"
       *   eg.
       * ( changenumber-oid NAME 'changenumber' DESC 'Standard Attribute' SYNTAX '1.3.6.1.4.1.1466.115.121.1.27' SINGLE-VALUE )
       */

      if (nv->get(CONFIG_ATTRIBUTETYPES))
      {
         strncpy(curEntryValueP, "( ", curValLen);
		 curLen = strlen(entryValue);
		 curValLen = entryValueLen - curLen;
		 curEntryValueP = entryValue + curLen;

         if (s = nv->get("oid"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "%s ", s);
         }

         if (s = nv->get(CONFIG_ATTRIBUTETYPES))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "NAME '%s' ", s);
         }

         if (s = nv->get("syntax"))
         {
            const char *oid;
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "SYNTAX '%s' ", (oid = syntaxToOID(s)) ? oid : s);
         }

         if (s = nv->get("single"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "SINGLE-VALUE  ", "");
         }
         snprintf(curEntryValueP, curValLen, ")");
      }
      return True;
   }
   else if (strcasecmp(type, SCHEMA_OBJECTCLASSES) == 0)
   {
      //  ( oid NAME '%s' DESC ' ' SUP '%s' MUST ( .. $ .. $ .. ) MAY ( .. $ .. $ .. ) )
      // eg. ( 2.5.6.0 NAME 'top' DESC 'Standard ObjectClass' MUST ( objectclass )  MAY ( aci ) )

      if (nv->get(CONFIG_OBJECTCLASSES))
      {
         const char *s;
         strncpy(curEntryValueP, "( ", curValLen);
		 curLen = strlen(entryValue);
		 curValLen = entryValueLen - curLen;
		 curEntryValueP = entryValue + curLen;

         if (s = nv->get("oid"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "%s ", s);
         }
         else
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "%s-oid ", nv->get(CONFIG_OBJECTCLASSES));
         }

         if (s = nv->get(CONFIG_OBJECTCLASSES))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "NAME '%s' ", s);
         }

         if (s = nv->get("superior"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "SUP '%s' ", s);
         }

         if (s = nv->get("requires"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "MUST ( %s ) ", s);
         }

         if (s = nv->get("allows"))
         {
            ADDFORMVALUE(entryValue, curEntryValueP, curValLen, "MAY ( %s ) ", s);
         }
         
         snprintf(curEntryValueP, curValLen, ")");
      }
      return True;
   }

   return False;
}


/***********************************************
 * Retrieve   oid=...
 *            name=...
 *            desc=...
 *            syntax=...
 *            alias=...
 * attribute attribute-name [attribute-aliases] attribute-oid syntax [single]
 * attribute attribute-name syntax [single]
 * eg. attribute aliasedObjectName    2.5.4.1     dn 
 */
Bool
LdapSchema::nextEntry(char *type, int typeLen, char *name, int nameLen,
                      char *entryValue, int entryValueLen)
{
   char buffer[BUFSIZ];
   char *line = NULL;
   char lineBuf[BUFSIZ];
   char *ret;

   /*
    *  Skip comment and blank lines
    */
   while (ret = fgets( buffer, sizeof(buffer), _schemaFile) )
   {
      buffer[strlen(buffer)-1] = 0;

      if ( buffer[0] == '#' )
      {
         // Comments - skip
         continue;
      }
      else if (isBlank(buffer) == True)
      {
         // skip blank lines
      }
      else
      {
         break;
      }
   }

   if (ret == NULL)
      return False;

   *entryValue = '\0';

   strncpy(lineBuf, buffer, sizeof(lineBuf));
   line = lineBuf;
   stripBlank(line, True);

   if (strncasecmp(line, CONFIG_ATTRIBUTETYPES, strlen(CONFIG_ATTRIBUTETYPES)) == 0)
   {
      char *p, *q;
      int i, n;
      char *item;
      char *elements[10];
      int aliasBegin, aliasEnd;
      NVPair nv;

      // change name from "attribute" to "attributetypes"
      strncpy(type, SCHEMA_ATTRIBUTETYPES, typeLen);

      p = line;
      for (i = 0; i < 10; i++)
      {
         item = p;
         if ((q = strpbrk(p, " \t")) != NULL)
         {
            *q = '\0';
            elements[i] = item;
            p = q + 1;
         }
         else {
            elements[i] = p;
            break;
         }
         stripBlank(p, False);
      }

      if (i == 10) {
         printf("Too much element to define attribute\n");
         return False;
      }

      n = i + 1;
      if (n < 3)
      {
         printf("Parsing error on line: %s\n", buffer);
         return False;
      }

      nv.set(elements[0], elements[1]);

      aliasBegin = 2;

      if (strcasecmp(elements[n-1], "single") == 0)
      {
         nv.set("single", "single");
         nv.set("syntax", elements[n - 2]);
         nv.set("oid", elements[n-3]);
         aliasEnd = n - 4;
      }
      else
      {
         nv.set("syntax", elements[n-1]);
         nv.set("oid", elements[n-2]);
         aliasEnd = n - 3;
      }

      // if oid is not numeric number, create netscape oid, name-oid
      const char *s;
      s = nv.get("oid");
      for (i = 0; (i < strlen(s)) && isdigit(s[i]); i++);
      if (i < strlen(s))
      {
         char oid[128];
         strncpy(oid, elements[1], sizeof(oid));
         strcat(oid, "-oid");
         nv.set("oid", oid);
      }

      if (aliasEnd >= aliasBegin)
         nv.set("alias", elements[aliasBegin]);

      if (aliasEnd > aliasBegin)
      {
         // Do we want to support multi aliases ???? 
         printf("Multi alias ???? (%d-%d) : %s\n", aliasBegin, aliasEnd, buffer);
      }

      formValue(type, &nv, entryValue, entryValueLen);
      strncpy(name, nv.get(CONFIG_ATTRIBUTETYPES), nameLen);

      // TAKE CARE OF alias HERE!!!!

   }
   else if (strncasecmp(line, CONFIG_OBJECTCLASSES, strlen(CONFIG_OBJECTCLASSES)) == 0)
   {
      char *seps = (char *)" \t";
      NVPair nv;
      Bool gotID = False;
      char *token;
      char value[2048];
      char id[128];

      // change from "objectclass" to "objectclasses"
      strncpy(type, SCHEMA_OBJECTCLASSES, typeLen);

      while (line) 
      {
         int vallen;
         token = strtok(line, seps);

         do 
         {
            if (token == NULL)
               break;

            if (gotID == True)
            {
               if (token[strlen(token) - 1] == ',')
               {
                  gotID = True;
                  token [strlen(token) - 1] = '\0';
               }
               else
               {
                  gotID = False;
               }

               if (value[0] != '\0')
               {
                  vallen = sizeof(value) - strlen(value);
                  strncat(value, " $ ", vallen);
               }
               vallen = sizeof(value) - strlen(value);
               strncat(value, token, vallen);

               nv.set(id, value);
            }
            else  
            {
               strncpy(id, token, sizeof(id));
               value[0] = '\0';
               gotID = True;
            }            
         }
         while ((token = strtok(NULL, seps)) != NULL);

         while ((ret = fgets(buffer, sizeof(buffer), _schemaFile)) != NULL)
         {
            if (buffer[strlen(buffer)-1] == '\n')
                buffer[strlen(buffer)-1] = '\0';

            if (buffer[0] != '#')
               break;
         }

         if (ret == NULL || isBlank(buffer))
         {
            break;
         }

         strncpy(lineBuf, buffer, sizeof(lineBuf));
         line = lineBuf;
      }      

      formValue(type,&nv,entryValue, entryValueLen);
      strncpy(name, nv.get(CONFIG_OBJECTCLASSES), nameLen);
   }

   if (ret == NULL && entryValue[0] == '\0')
      return False;
   else
      return True;
}

#define STANDARD_ATTRIBUTE "DESC 'Standard Attribute'"
#define STANDARD_OBJECTCLASS "DESC 'Standard ObjectClass'"
#define USER_DEFINED_SCHEMA_5X "X-ORIGIN 'user defined'"

Bool LdapSchema::isStandardSchema(const char *value)
{

   if (strstr(value, USER_DEFINED_SCHEMA_5X))
      return False;

   if (strstr(value, STANDARD_ATTRIBUTE) || strstr(value, STANDARD_OBJECTCLASS))
      return True;
   else
      return False;     
}


int LdapSchema::updateEntryValue(char *type, char *name, char *value, int doUpdate)
{
   char *values[2];
   LDAPMod *mods[3];
   NVPair *nv;
   int err = OKAY;

   values[0] = value;
   values[1] = NULL;

   if (strcasecmp(type, SCHEMA_OBJECTCLASSES) == 0)
      nv = & _oc;
   else if (strcasecmp(type, SCHEMA_ATTRIBUTETYPES) == 0)
      nv = & _at;
   else
      return 1;  // operation error


//   if (nv->get(name) && doUpdate)
   if (nv->get(name) && doUpdate && (isStandardSchema(nv->get(name)) == False))
   {  
      LDAPMod mod1, mod2;
      char *prevValues[2];
      char *prevValue;

      prevValue = strdup(nv->get(name));
      prevValues[0] = prevValue;
      prevValues[1] = NULL;

      // update
      mod1.mod_op = LDAP_MOD_DELETE ;
      mod1.mod_type = type;
      mod1.mod_values = prevValues;

      mod2.mod_op = LDAP_MOD_ADD;
      mod2.mod_type = type;
      mod2.mod_values = values;

      mods[0] = &mod1;
      mods[1] = &mod2;
      mods[2] = NULL;

      err = domodify(mods);

      if (err != OKAY)
      {
         setupLogMessage(SETUP_LOG_FATAL, "schema", "updating %s: %s,%s", type, mod2.mod_type, value);
      }

   }
   else
   {
      // append
      LDAPMod mod;

      mod.mod_op = LDAP_MOD_ADD;
      mod.mod_type = type;
      mod.mod_values = values;

      mods[0] = &mod;
      mods[1] = NULL;
     
      err = domodify(mods);

      if (err != OKAY)
      {
         setupLogMessage(SETUP_LOG_FATAL, "schema", "writing new %s: %s,%s", type, mod.mod_type, value);
      }
   }

   return err;

}


int LdapSchema::insertSchemaFile(const char *fileName, int doUpdate)
{
   char type[128], name[128], value[2048];
   int err = OKAY;

   _schemaFile = fopen (fileName, "r");

   if (_schemaFile == NULL) {
//      printf("Fail: open schema definition file\n");
      return 1;
   }

   if ((err = retrieve()) == OKAY)
   {
      while (nextEntry(type, sizeof(type), name, sizeof(name), value, sizeof(value)) == True)
      {
         err = updateEntryValue(type, name, value, doUpdate);

//         sleep(1);
/*
         if (err != OKAY)
         {
            return err;
         }
*/
      }
   }
  
   fclose(_schemaFile);

   return err;
}



extern "C"
{
NSAPI_EXTERN 
LdapErrorCode insertSchemaEntries (Ldap *ldap, const char *schemaFile)
{
   LdapSchema ldapSchema(ldap);
   int doUpdate = 1;

   if (ldap == NULL || schemaFile == NULL)
   {
      return INVALID_INPUT;
   }

   return ldapSchema.insertSchemaFile(schemaFile, doUpdate);
}

NSAPI_EXTERN 
LdapErrorCode appendSchemaEntries (Ldap *ldap, const char *schemaFile)
{
   LdapSchema ldapSchema(ldap);
   int doUpdate = 0;

   if (ldap == NULL || schemaFile == NULL)
   {
      return INVALID_INPUT;
   }

   return ldapSchema.insertSchemaFile(schemaFile, doUpdate);
}


} /* End extern "C" */
