/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version
 * 2.1 of the License.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * END COPYRIGHT BLOCK **/
extern "C" {
#include <stdlib.h>
#include <stdio.h>
#if defined(AIX)
   #include <strings.h>
#endif
#include <string.h>
#include <ctype.h>
}

#include "global.h"
#include "ldapu.h"
#include "misc.h"
#include "nsutils.h"
#include "setupapi.h"
#include "setupdefs.h"
#include "setupldap.h"

static LdapErrorCode
walkUserDirChain (LdapEntry & ldapEntry, const char *entryDN)
{
   LdapErrorCode err = NOT_FOUND;
   const char *s;

   for (int i = 0; i < 3; i++)
   {
      err = ldapEntry.retrieve(entryDN);

      if (err == OKAY)
      {
     if (s = ldapEntry.getAttribute("nsDirectoryURL"))
       // Find it
       break;
         else
         {
            s = ldapEntry.getAttribute("nsDirectoryInfoRef");
            if (s)
            {
               entryDN = s;
            }
            else
            {
               err = NOT_FOUND;
               break;
            }
         }
      }
   }
   return err;
}

static LdapErrorCode
setupGlobalPreferences(const Ldap *ldap, const char *adminDomain)
{
   LdapEntry ldapEntry(ldap);
   LdapError err;
   NSString globalPref = DEFAULT_GLOBAL_PREFS_RDN;
   char * domain = formAdminDomainDN(adminDomain);

   NSString defaultDescription = "Default branch for ";
   defaultDescription = defaultDescription + SETUP_DEFAULTNAME + " Global Preferences";

   globalPref = globalPref + LDAP_PATHSEP + domain;

   if (ldapEntry.retrieve(globalPref) != OKAY)
   {
      ldapEntry.setAttribute("objectclass", DEFAULT_GLOBAL_PREFS_OBJECT);
      ldapEntry.setAttribute("ou", DEFAULT_GLOBAL_PREFS);
      ldapEntry.setAttribute("aci", DEFAULT_GLOBAL_PREFS_ACI);
      ldapEntry.setAttribute("description", defaultDescription);
      err = ldapEntry.insert(globalPref);
   }
   else
   {
      ldapEntry.setAttribute("aci", DEFAULT_GLOBAL_PREFS_ACI);
      ldapEntry.setAttribute("description", defaultDescription);
      err = ldapEntry.replace(globalPref);
   }

   if(domain) free(domain);

   return err;
}

static char *
formAdminGroupSearchCriteria(const char *installedLocation)
{
   char *criteria;

   if (installedLocation == NULL)
   {
      return NULL;
   }

#ifdef XP_WIN32
{
   char *lpszValue;
   lpszValue = (char *)installedLocation;
   CHAR szTargetDir[260];
   int i= 0;
   while (*lpszValue)
   {
      if (*lpszValue == '\\')
      {
         szTargetDir[i++] = '\\';
         szTargetDir[i++] = '\\';
      }
      else
      {
         szTargetDir[i++] = *lpszValue;
      }
      lpszValue++;
   }
   szTargetDir[i] = '\0';
   criteria = strdup(NSString("(&(objectclass=nsAdminGroup)(nsConfigRoot=") + szTargetDir + "))");
}
#else
  criteria = strdup(NSString("(&(objectclass=nsAdminGroup)(nsConfigRoot=") + installedLocation + "))");
#endif

   return criteria;

}

static Bool
testReplicationDeleteStatus (const Ldap *ldap, const char *entryDN)
{
   Bool status = False;

   if (entryDN == NULL)
   {
      return True; // Ignore
   }

   LdapEntry theEntry(ldap);

   for (int i = 0; i < 5 && status == False; i++)
   {
      status = theEntry.exists(entryDN) ? False : True;
      if (status == False)
      {
         WAIT_A_BIT(5);
      }
   }
   return status;
}

static Bool
testReplicationUpdateStatus (const Ldap *ldap, const char *entryDN, const char *attr, const char *value)
{
   Bool status = False;

   if (ldap == NULL || entryDN == NULL)
   {
      status = True; // Ignore
   }
   else
   {
      LdapEntry theEntry(ldap);

      for (int i = 0; i < 5 && status == False; i++)
      {
         if (attr == NULL ||  value == NULL)
         {
            status = theEntry.exists(entryDN);
         }
         else
         {
            LdapError err = theEntry.retrieve(entryDN);
            if (err == OKAY)
            {
               status = theEntry.checkAttribute(attr, value) ? True : False;
            }
            else
            {
               status = False;
            }
         }

         if (status == False)
         {
            WAIT_A_BIT(5);
         }

      }
   }

   return status;
}


NSAPI_EXTERN LdapErrorCode setupCreateLdap (Ldap **ldap,
                                   const char *ldapURL,
                                   const char *ldapUser,
                                   const char *ldapPwd,
                                   const char *binddn,
                                   const char *bindpwd)
{
   LdapError err;

   if (ldap == NULL || ldapURL == NULL )
   {
      return INVALID_INPUT;
   }

   Ldap *ldapObj = new Ldap( err, ldapURL, ldapUser, ldapPwd, binddn, bindpwd);

   if (err != OKAY)
   {
      delete ldapObj;
      *ldap = NULL;
   }
   else
   {
      *ldap = ldapObj;
   }
   return err;
}

NSAPI_EXTERN void setupDestroyLdap(Ldap *ldap)
{
   if (ldap)
      delete ldap;
}

NSAPI_EXTERN const char *setupGetUserDN (const Ldap *ldap)
{
   if (ldap == NULL) return NULL;
   return ldap->userDN();
}

NSAPI_EXTERN const char *setupGetServerVersion (const Ldap *ldap)
{
   if (ldap == NULL) return NULL;
   return ldap->version();
}

NSAPI_EXTERN LdapEntry *setupCreateLdapEntry (const Ldap *ldap)
{
   LdapEntry *ldapEntry = NULL;
   if (ldap)
   {
      ldapEntry = new LdapEntry(ldap);
   }
   return ldapEntry;
}

NSAPI_EXTERN void setupDestroyLdapEntry(LdapEntry *ldapEntry)
{
   if (ldapEntry)
      delete ldapEntry;
}

NSAPI_EXTERN const char *setupGetEntryDN (const LdapEntry *ldapEntry)
{
   if (ldapEntry)
   {
      return ldapEntry->entryDN();
   }
   else
   {
      return NULL;
   }
}

/* DS50 and higher requires a unique name for each suffix. Use
 * setupGenBackend with a prefix and it'll give you a unique name. Eg:
 * give it: "MyServer" and it might return something like: "MyServer15"
 * which you can then pass to setupCreateSuffixAndBackend*/
NSAPI_EXTERN char * setupGenBackendName(Ldap *ldap, char *suffix, char *prefix) {
  LdapEntry *beEntries = new LdapEntry(ldap);
  char *filter;
  int filterlen;
  char **explodedDN;
  int beMax=0;
  int beCurr;
  char *beNumPtr;
  char *name;
  int namelen;
  int err;

  filterlen = (7 + strlen(prefix)) * sizeof(char);
  filter = (char *) malloc(filterlen);
  if (filter == NULL) {
    return(NULL);
  }
  snprintf(filter, filterlen, "(cn=%s*)", prefix);
  
  err = beEntries->retrieve(filter, LDAP_SCOPE_ONELEVEL, "cn=ldbm database,cn=plugins,cn=config");
  free(filter);

  while (err == OKAY) {
    explodedDN = ldap_explode_dn((char *)beEntries->entryDN(), 1);
    beNumPtr = explodedDN[0] + strlen(prefix);
    beCurr = atoi(beNumPtr);
    if (beCurr > beMax) {
      beMax = beCurr;
    }
    err = beEntries->next();
  } 
  setupDestroyLdapEntry(beEntries);

  beMax++;

  namelen = (35 + strlen(prefix)) * sizeof(char);
  name = (char *) malloc(namelen);
  if (name == NULL) {
    return(NULL);
  }
  snprintf(name, namelen, "%s%d", prefix, beMax);

  return(name);
}

/* Use setupCreateSuffixAndBackend for DS5.0 and higher. Use
 * genBackendName() if you want to guarantee a unique backend name (for the name argument) */
NSAPI_EXTERN LdapErrorCode setupCreateSuffixAndBackend(Ldap *ldap, char *suffix, char *name)
{
  LdapEntry *beEntry;
  LdapEntry *mapEntry;
  char *beDN;
  int dnlen;
  char *mapDN;
  int mapdnlen;
  char *enqSuff;
  int suffixlen;
  LdapErrorCode rez;
  
  if (ldap == NULL || suffix == NULL || name == NULL ||
      *suffix == '\0' || *name == '\0') {
    return False;
  }

  /*The backend entry DN*/
  dnlen = (42 + strlen(name)) * sizeof(char);
  beDN = (char *) malloc(dnlen);
  if (beDN == NULL) {
    return False;
  }
  snprintf(beDN, dnlen, "cn=%s,cn=ldbm database,cn=plugins,cn=config", name);

  /*The backend entry*/
  beEntry = new LdapEntry(ldap);
  beEntry->addAttribute("objectclass","top");
  beEntry->addAttribute("objectclass","nsBackendInstance");
  beEntry->addAttribute("objectclass","extensibleObject");
  beEntry->addAttribute("cn",name);
  beEntry->addAttribute("nsslapd-suffix",suffix);
  beEntry->addAttribute("nsslapd-cachesize", "-1");
  beEntry->addAttribute("nsslapd-cachememsize", "10485760");

  /* Enquote the suffix */
  suffixlen = (3 + strlen(suffix)) * sizeof(char);
  enqSuff = (char *) malloc(suffixlen);
  if (enqSuff == NULL) {
    return False;
  }
  snprintf(enqSuff, suffixlen, "\"%s\"", suffix);

  /* The mapping tree entry DN */
  mapdnlen = (32 + strlen(suffix)) * sizeof(char);
  mapDN = (char *) malloc(mapdnlen);
  if (mapDN == NULL) {
    return False;
  }
  snprintf(mapDN, mapdnlen, "cn=\"%s\",cn=mapping tree,cn=config", suffix);

  /* The mapping tree entry */
  mapEntry = new LdapEntry(ldap);
  mapEntry->addAttribute("cn",enqSuff);
  mapEntry->addAttribute("objectclass","top");
  mapEntry->addAttribute("objectclass","extensibleObject");
  mapEntry->addAttribute("objectclass","nsMappingTree");
  mapEntry->addAttribute("nsslapd-state", "backend");
  mapEntry->addAttribute("nsslapd-backend", name);

  rez = beEntry->insert(beDN);
  
  if (rez == LDAP_SUCCESS){
    rez = mapEntry->insert(mapDN);
  }

  setupDestroyLdapEntry(beEntry);
  setupDestroyLdapEntry(mapEntry);
  free(beDN);
  free(mapDN);
  free(enqSuff);

  return rez;
}



NSAPI_EXTERN LdapErrorCode setupCreateSuffix(Ldap *ldap, char *suffix)
{
#define ENTRY_DN "cn=config,cn=ldbm"
#define ATTRIBUTE_NAME "nsslapd-suffix"

   LDAPMod mod1;
   char *values[2];
   LDAPMod *mods[2];

  if (ldap == NULL || suffix == NULL)
      return False;
  values[0] = suffix;
  values[1] = NULL;

  mod1.mod_op = LDAP_MOD_ADD;
  mod1.mod_type = (char *)ATTRIBUTE_NAME;
  mod1.mod_values = values;

  mods[0] = &mod1;
  mods[1] = NULL;

  return ldap->modify(ENTRY_DN, mods);
}


NSAPI_EXTERN Bool setupEntryExists(LdapEntry *ldapEntry, const char *entryDN)
{
   if (ldapEntry == NULL || entryDN == NULL)
      return False;
   else
      return ldapEntry->exists(entryDN);
}

NSAPI_EXTERN LdapErrorCode setupRetrieveLdapEntry(LdapEntry *ldapEntry, const char *entryDN)
{
   if (ldapEntry == NULL || entryDN == NULL)
      return INVALID_INPUT;
   else
      return ldapEntry->retrieve(entryDN);
}

NSAPI_EXTERN LdapErrorCode setupRetrieveLdapEntries(LdapEntry *ldapEntry,
      const char *criteria, int scope, const char *baseDN)
{
   if (ldapEntry == NULL || criteria == NULL)
   {
      return INVALID_INPUT;
   }
   else
   {
      return ldapEntry->retrieve(criteria, scope, baseDN);
   }
}

NSAPI_EXTERN LdapErrorCode setupNextLdapEntry(LdapEntry *ldapEntry)
{
   if (ldapEntry == NULL)
   {
      return INVALID_INPUT;
   }
   else
   {
      return ldapEntry->next();
   }
}

NSAPI_EXTERN char **setupGetLdapAttributeNames(const LdapEntry *ldapEntry)
{
   if (ldapEntry == NULL)
   {
      return NULL;
   }

   return ldapEntry->getAttributeNames();
}

NSAPI_EXTERN void setupFreeLdapAttributeNames(LdapEntry *ldapEntry, char **attrs)
{
   if (ldapEntry == NULL || attrs == NULL)
   {
      return;
   }

   ldapEntry->freeAttributeNames(attrs);
}

NSAPI_EXTERN char **setupGetLdapAttributes(const LdapEntry *ldapEntry, const char *name)
{
   if (ldapEntry == NULL || name == NULL)
      return NULL;

   return ldapEntry->getAttributes(name);
}

NSAPI_EXTERN void setupFreeLdapAttributes(LdapEntry *ldapEntry, char **attrs)
{
   ldapEntry->freeAttributes(attrs);
}

NSAPI_EXTERN void setupSetLdapAttribute(LdapEntry *ldapEntry, const char *name, const char *value)
{
   if (ldapEntry == NULL || name == NULL || value == NULL)
   {
      return;
   }
   else
   {
      ldapEntry->setAttribute(name,value);
   }
}

NSAPI_EXTERN void setupAddLdapAttribute (LdapEntry *ldapEntry, const char *name, const char *value)
{
   if (ldapEntry == NULL || name == NULL || value == NULL)
      return;
   else
   {
      ldapEntry->addAttribute(name, value);
   }
}

NSAPI_EXTERN void setupAddLdapAttributes(LdapEntry *ldapEntry, const char *name, const char **values)
{
   if (ldapEntry == NULL || name == NULL || values == NULL)
   {
      return;
   }
   else
   {
      for (int i = 0; values[i] != NULL && i < 500; i++) // Guard
      {
        ldapEntry->addAttribute(name, values[i]);
      }
   }
}


NSAPI_EXTERN LdapErrorCode setupRenameLdapEntry(LdapEntry *ldapEntry, const char *entryDN, const char *rdn)
{
   if (ldapEntry == NULL || entryDN == NULL || rdn == NULL)
   {
      return INVALID_INPUT;
   }
   else
   {
      return ldapEntry->rename(entryDN, rdn);
   }
}

NSAPI_EXTERN LdapErrorCode setupReplaceLdapEntry(LdapEntry *ldapEntry, const char *entryDN)
{
   if (ldapEntry == NULL || entryDN == NULL)
   {
      return INVALID_INPUT;
   }
   else
   {
      return ldapEntry->replace(entryDN);
   }
}

NSAPI_EXTERN LdapErrorCode setupInsertLdapEntry(LdapEntry *ldapEntry, const char *entryDN)
{
   if (ldapEntry == NULL || entryDN == NULL)
      return INVALID_INPUT;
   else
      return ldapEntry->insert(entryDN);
}

NSAPI_EXTERN LdapErrorCode setupUpdateLdapEntry(LdapEntry *ldapEntry, const char *entryDN)
{
   if (ldapEntry == NULL || entryDN == NULL)
      return INVALID_INPUT;
   else
      return ldapEntry->update(entryDN);
}

NSAPI_EXTERN LdapErrorCode setupDropLdapEntry (Ldap *ldap, const char *entryDN)
{
   if (ldap == NULL || entryDN == NULL)
      return INVALID_INPUT;
   else
   {
      LdapEntry ldapEntry(ldap);
      return ldapEntry.drop(entryDN);
   }
}

NSAPI_EXTERN LdapErrorCode setupDropAllLdapEntries (Ldap *ldap, const char *entryDN)
{
   if (ldap == NULL || entryDN == NULL)
      return INVALID_INPUT;
   else
   {
      LdapEntry ldapEntry(ldap);
      return ldapEntry.dropAll(entryDN);
   }
}

NSAPI_EXTERN void setupPrintLdapEntry (const LdapEntry *ldapEntry, const char *fn)
{
   if (ldapEntry)
   {
      ldapEntry->printEntry(fn);
   }
}

/*********************************************************************
**
** FUNCTION:    verifyLdapUrl
** DESCRIPTION:
**   Verify that an LDAP URL is valid
**
** INPUTS:      url  - the ldap url.
** OUTPUTS:     none
** RETURN:
**   Refer to error codes defined in ldapu.h
** SIDE EFFECTS:
**      None
** RESTRICTIONS:
**      None
** MEMORY:
**********************************************************************
*/
NSAPI_EXTERN  LdapErrorCode
setupVerifyLdapUrl(const char *url, const char *bindDN, const char *bindPwd)
{
   LdapError err;

   Ldap ldap(err, url, bindDN, bindPwd, NULL, NULL);

   return err;
}


/*********************************************************************
**
** FUNCTION:    authLdapUser
** DESCRIPTION:
**   One-shot authentication of an LDAP user.
**
** INPUTS:
**   url  - the ldap url.
**   ldapUser - the user
**   ldapPwd  - the password
**   binddn, bindpwd - optional DN and Password to bind as
*
** OUTPUTS:     none

**   0 - Good
**   Other - failed (refer to ldapu.h)
**
** SIDE EFFECTS:
**      None
** RESTRICTIONS:
**      None
** MEMORY:
**********************************************************************
*/
NSAPI_EXTERN  LdapErrorCode
setupAuthLdapUser(const char *url,
               const char *ldapUser, const char *ldapPwd,
               const char *binddn, const char *bindpw)
{
   LdapError err;

   Ldap ldap(err, url, ldapUser, ldapPwd, binddn, bindpw);

   return err;
}


NSAPI_EXTERN LdapErrorCode setupCreateSIE(
     LdapEntry *sieEntry,
         LdapEntry *appEntry,
         const char *hostName,
         const char *installedLocation,
         const char *adminDomain)
{
   LdapError err;
   char **objClassList;
   char *appName;
   char *appVersion;
   char aci[BIG_BUF];
   char tmp[MED_BUF];
   NSString   instanceName;
   NSString sieDN;
   NSString isieDN;
   Bool isVersion4 = False;
   NVPair tm;

   // cn = appName, cn= admin group, cn = hostname, cn = netscapeRoot

   if (sieEntry == NULL || appEntry == NULL || hostName == NULL || installedLocation == NULL || adminDomain == NULL)
   {
      return INVALID_INPUT;
   }
   else if (sieEntry->isEmpty() || appEntry->isEmpty())
   {
      return INVALID_OBJECT;
   }
   else if (sieEntry->checkAttribute("objectclass", "netscapeServer") == False ||
       sieEntry->getAttribute("nsServerID") == NULL)
   {
      return MISSING_ATTR;
   }

   if (appEntry->ldap()->version()[0] == '4')
   {
      isVersion4 = True;
   }

   objClassList = appEntry->getAttributes("objectclass");

   if ((appName = appEntry->getAttribute("cn")) == NULL)
   {
      appName = appEntry->getAttribute("nsProductName");
      appEntry->setAttribute("cn", appName);
   }

   appVersion = appEntry->getAttribute("nsProductVersion");

   if (objClassList == NULL || appName == NULL || appVersion == NULL)
   {
      if (objClassList != NULL)
      {
         free(objClassList);
      }
      return INVALID_OBJECT;
   }

   free (appVersion);

   //
   // Check if o=NetscapeRoot exists
   //

   char *criteria;

   LdapEntry rootEntry(appEntry->ldap());

   if (rootEntry.exists(DEFAULT_ROOT_DN) == False)
   {
      return INVALID_ROOT;
   }
   else if (isVersion4)
   {
      // We might fail here, but the first time ever should
      // always succeed
      // DS installation already set this up so we don't have to
      //      rootEntry.setAttribute("aci", DEFAULT_ROOT_ACI_3);
      //     rootEntry.update(DEFAULT_ROOT_DN);
   }

   //
   // Check if adminDomain exists
   //

   NSString domain = formAdminDomainDN(adminDomain);

   if (rootEntry.retrieve (domain) != OKAY ||
       setupGlobalPreferences(appEntry->ldap(), domain) != OKAY)
   {
      return INSTALL_DN_ERR;
   }
   else
   {
      snprintf(aci, sizeof(aci), DEFAULT_SS_CONTAINER_ACI,
                     rootEntry.getAttribute("ou"));
      if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
      rootEntry.setAttribute("aci", aci);

      // We'll update later on if we add a new host
   }

   //
   // Now look for my host under domain
   //
   LdapEntry hostEntry = LdapEntry(appEntry->ldap());
   NSString hostDN;
   hostDN = NSString("cn=") + hostName + LDAP_PATHSEP + domain;

   char *sHostName = strdup(hostName);
   char *s = strchr(sHostName, '.');
   if  (s)
   {
      *s = 0;
   }

   if (hostEntry.retrieve(hostDN) != OKAY)
   {
      SysInfo sysInfo;
      hostEntry.addAttribute("objectclass", "top");
      hostEntry.addAttribute("objectclass", "nsHost");
      hostEntry.addAttribute("objectclass", GROUP_OBJECT_CLASS);
      hostEntry.addAttribute("cn", hostName);
      hostEntry.addAttribute("serverHostname", hostName);
      hostEntry.addAttribute("nsOsVersion", sysInfo.get("osVersion"));
      hostEntry.addAttribute("nsHardwarePlatform", sysInfo.get("hardwarePlatform"));

      if (isVersion4 == False)
      {
         snprintf(aci, sizeof(aci), DEFAULT_HOST_ACI, hostDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         hostEntry.addAttribute("aci", aci);
      }

      err = hostEntry.insert(hostDN);
      if (err == LDAP_ALREADY_EXISTS) err = LDAP_SUCCESS; //GWB hack

      if (err)
      {
         return err;
      }
      else
      {
         rootEntry.replace(domain);
      }
   }

   testReplicationUpdateStatus(hostEntry.ldap(), hostEntry.entryDN(), NULL, NULL);

   LdapEntry groupEntry = LdapEntry(appEntry->ldap());
   NSString groupDN;
   NSString adminServerName;
   NSString adminSieDN;

   // Search for the Admin Group associated with the
   // installedLocation - hostDN is the DN of the Host node at this
   // time
   criteria = formAdminGroupSearchCriteria(installedLocation);
   err = groupEntry.retrieve(criteria, LDAP_SCOPE_ONELEVEL, hostDN);
   free(criteria);

   adminServerName = NSString(SETUP_DEFAULTBRAND) + " Administration Server";

   if (err == OKAY)
   {
      groupDN = groupEntry.entryDN();
      adminSieDN = NSString("cn=admin-serv-") + sHostName + LDAP_PATHSEP + "cn="
              + adminServerName + LDAP_PATHSEP + groupDN;

      snprintf(aci, sizeof(aci), DEFAULT_ADMINGRP4_ACI, adminSieDN.data());
      if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
      groupEntry.addAttribute("aci", aci);

      if (isVersion4 == False)
      {
         snprintf(aci, sizeof(aci), DEFAULT_ADMINGRP_ACI, groupDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         groupEntry.addAttribute("aci", aci);
      }

      groupEntry.replace(groupDN);

   }
   else
   {
      int count = 0;
      NSString groupName = DEFAULT_ADMIN_GROUP;

      groupDN = NSString("cn=") + groupName + LDAP_PATHSEP + hostDN;

      // Search for a free one
      while (groupEntry.exists(groupDN) == True)
      {
         if (count == 0) ++count;
         snprintf(tmp, sizeof(tmp), "(%d)", ++count);
         if (!memchr(tmp, 0, sizeof(tmp))) return INSTALL_DN_ERR; /*overflow*/
         groupName = NSString(DEFAULT_ADMIN_GROUP) + " " + tmp;
         groupDN = NSString("cn=") + groupName + LDAP_PATHSEP + hostDN;
      }

      groupEntry.clear();

      groupEntry.addAttribute("objectclass", DEFAULT_ADMIN_GROUP_OBJECT);
      groupEntry.addAttribute("objectclass", GROUP_OBJECT_CLASS);
      groupEntry.addAttribute("nsAdminGroupName", groupName);
      groupEntry.addAttribute("nsConfigRoot", installedLocation);

      NSString userDirRef(DEFAULT_USER_DIRINFO_RDN);
      groupEntry.addAttribute("objectclass", DEFAULT_USER_DIRINFO_OBJECT);
      userDirRef = userDirRef + LDAP_PATHSEP + DEFAULT_GLOBAL_PREFS_RDN + LDAP_PATHSEP + domain;

      groupEntry.setAttribute("nsDirectoryInfoRef", userDirRef);

      adminSieDN = NSString("cn=admin-serv-") + sHostName + LDAP_PATHSEP + "cn="
              + adminServerName + LDAP_PATHSEP + groupDN;
      groupEntry.setAttribute("nsAdminSIEDN", adminSieDN);

      if (isVersion4 == False)
      {
         snprintf(aci, sizeof(aci), DEFAULT_ADMINGRP_ACI, groupDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         groupEntry.addAttribute("aci", aci);
         snprintf(aci, sizeof(aci), DEFAULT_ADMINGRP4_ACI, adminSieDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         groupEntry.addAttribute("aci", aci);
      }
      else if (strcasecmp(appName, adminServerName))
      {
         snprintf(aci, sizeof(aci), DEFAULT_ADMINGRP4_ACI, adminSieDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         groupEntry.addAttribute("aci", aci);
      }

      err = groupEntry.insert(groupDN);
      if (err == LDAP_ALREADY_EXISTS) err = LDAP_SUCCESS; //GWB hack
      if (err != OKAY)
      {
         return err;
      }

      testReplicationUpdateStatus(groupEntry.ldap(), groupEntry.entryDN(), NULL, NULL);

      hostEntry.addAttribute("uniquemember", groupDN);
      hostEntry.update(hostDN);
   }

   // Now it's turn for appEntry
   // All entries are supposed to be set already

   NSString appDN;
   appDN = NSString("cn=") + appName + LDAP_PATHSEP + groupDN;

   // Make sure these are set
   tm.stampTime();
   appEntry->addAttribute("objectclass", GROUP_OBJECT_CLASS);
   appEntry->addAttribute("objectclass", "nsApplication");
   appEntry->addAttribute("installationTimeStamp", tm.timeStamp());

   LdapEntry tmpAppEntry(appEntry->ldap());

   if (tmpAppEntry.retrieve(appDN) != OKAY)
   {
      if (isVersion4 == False)
      {
         snprintf(aci, sizeof(aci), DEFAULT_PRODUCT_ACI, appDN.data());
         if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
         appEntry->addAttribute("aci", aci);
      }

      err = appEntry->insert(appDN);
      if (err == LDAP_ALREADY_EXISTS) err = LDAP_SUCCESS; //GWB hack
      if (err == OKAY)
      {
         groupEntry.addAttribute("uniquemember", appDN);
         groupEntry.update(groupDN);
      }
   }
   else
   {
      char **members = tmpAppEntry.getAttributes("uniquemember");
      char **acis = tmpAppEntry.getAttributes("aci");

      appEntry->addAttributes("uniquemember", (const char **) members);
      appEntry->addAttributes("aci", (const char **) acis);

      err = appEntry->replace(appDN);

      tmpAppEntry.freeAttributes(members);
      tmpAppEntry.freeAttributes(acis);

   }
        
   if (err == LDAP_ALREADY_EXISTS) err = LDAP_SUCCESS; //GWB hack
   if (err != OKAY)
   {
      return err;
   }

   testReplicationUpdateStatus(appEntry->ldap(), appEntry->entryDN(), "installationTimeStamp", tm.timeStamp());

   free (appName);

   isieDN = appEntry->entryDN();
   appEntry->retrieve(isieDN);   // Reload to pickup what was there

   // Make sure consistency here
   sieEntry->addAttribute("objectclass", GROUP_OBJECT_CLASS);
   sieEntry->setAttribute("serverHostName", hostName);

   instanceName = sieEntry->getAttribute("cn");
   if (instanceName == (char *) NULL)
   {
      instanceName = sieEntry->getAttribute("nsServerId");
      sieEntry->setAttribute("cn", instanceName);
   }

   // e.g. cn=https-lupine, ou=NetscapeServers, o=mcom.com
   sieDN = NSString("cn=") + instanceName + LDAP_PATHSEP + isieDN;

   snprintf(aci, sizeof(aci), DEFAULT_SIE_ACI_1, sieDN.data());
   if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
   sieEntry->addAttribute("aci", aci);
   snprintf(aci, sizeof(aci), DEFAULT_SIE_ACI_2, sieDN.data());
   if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
   sieEntry->addAttribute("aci", aci);
   /*
    * SIE is a member of its own group
    */
   sieEntry->addAttribute("uniquemember", sieDN);

   if (adminSieDN != (char *) NULL)
   {
      sieEntry->addAttribute("uniquemember", adminSieDN);
   }

   sieEntry->addAttribute("installationTimeStamp", tm.timeStamp());

   if (sieEntry->exists(sieDN) == False)
   {
      err = sieEntry->insert(sieDN);
   }
   else
   {
      err = sieEntry->replace(sieDN);
   }

   if (err == OKAY)
   {
      testReplicationUpdateStatus(sieEntry->ldap(), sieEntry->entryDN(), "installationTimeStamp", tm.timeStamp());
      appEntry->addAttribute("uniquemember", sieDN);
      appEntry->update(appDN);
   }

   return err;
}

NSAPI_EXTERN LdapErrorCode
setupUpdateISIEVersionInfo (Ldap *ldap, LdapEntry *appEntry,
              const char *productName, const char *hostName,
              const char *installedLocation, const char *adminDomain)
{
   if (ldap == NULL || appEntry == NULL || hostName == NULL ||
       installedLocation == NULL || adminDomain == NULL || productName == NULL)
   {
      return INVALID_INPUT;
   }
   else if (appEntry->isEmpty())
   {
      return INVALID_OBJECT;
   }

   NSString domain = formAdminDomainDN(adminDomain);
   LdapEntry domainEntry(ldap);

   if (domainEntry.retrieve(domain) != OKAY)
   {
      return INSTALL_DN_ERR;
   }

   LdapEntry hostEntry = LdapEntry(ldap);
   NSString hostDN;
   hostDN = NSString("cn=") + hostName + LDAP_PATHSEP + domain;

   if (hostEntry.retrieve(hostDN) != OKAY)
   {
      return OKAY;
   }

   LdapEntry groupEntry = LdapEntry(ldap);
   NSString groupDN;
   NSString adminSieDN;

   char *criteria = formAdminGroupSearchCriteria(installedLocation);

   if (groupEntry.retrieve(criteria, LDAP_SCOPE_ONELEVEL, hostDN) != OKAY)
   {
      return OKAY;
   }

   free(criteria);

   groupDN = groupEntry.entryDN();

   NSString appDN;
   appDN = NSString ("cn=") + productName + LDAP_PATHSEP + groupDN;

   LdapEntry tmpAppEntry(ldap);

   if (tmpAppEntry.retrieve(appDN) != OKAY)
   {
      return OKAY;
   }

   tmpAppEntry.setAttribute("nsVendor", appEntry->getAttribute("nsVendor"));
   tmpAppEntry.setAttribute("nsProductVersion", appEntry->getAttribute("nsProductVersion"));
   tmpAppEntry.setAttribute("nsBuildNumber", appEntry->getAttribute("nsBuildNumber"));
   tmpAppEntry.setAttribute("nsRevisionNumber", appEntry->getAttribute("nsRevisionNumber"));
   tmpAppEntry.setAttribute("nsSerialNumber", appEntry->getAttribute("nsSerialNumber"));
   tmpAppEntry.setAttribute("nsExpirationDate", appEntry->getAttribute("nsExpirationDate"));
   tmpAppEntry.setAttribute("nsBuildSecurity", appEntry->getAttribute("nsBuildSecurity"));
   tmpAppEntry.setAttribute("nsLdapSchemaVersion", appEntry->getAttribute("nsLdapSchemaVersion"));

   NVPair tm;
   tm.stampTime();
   tmpAppEntry.setAttribute("installationTimeStamp", tm.timeStamp());

   tmpAppEntry.replace(appDN);

   return OKAY;
}

NSAPI_EXTERN LdapErrorCode
setupRemoveSIE (Ldap *ldap, const char *sieDN, Bool deleteAll)
{
   LdapError err;
   LdapEntry entry(ldap);
   NSString sieRefDN;
   char *isieDN = NULL;

   if (entry.exists(sieDN) == False)
   {
      return INVALID_DN;
   }

   err = entry.dropAll(sieDN);

   if (err == OKAY)
   {
      isieDN = strchr((char *)sieDN, ',');
      if (isieDN)
      {
        ++isieDN;
      }
      else
      {
         return UNKNOWN_ERR;
      }

      if (deleteAll)
      {
         err = removeISIE(ldap, isieDN);
      }
   }

   return err;
}

NSAPI_EXTERN LdapErrorCode
setupRemoveISIE(Ldap *ldap, const char *isieDN)
{
   LdapError err;
   LdapEntry entry(ldap);
   char dn[MED_BUF];
   char *dnp;
   int dnlen = sizeof(dn);

   err = entry.retrieve(isieDN);

   if (err != OKAY)
   {
      return err;
   }

   // OK to remove, recursively go up the tree and remove all
   char **explodedDN = ldap_explode_dn(isieDN, 0);
   int i = 0;

   while (1)
   {
      char **s = &explodedDN[i];
      dn[0] = 0;
      dnp = &dn[0];
      while (*s != NULL)
      {
         int curlen = strlen(*s) + strlen(LDAP_PATHSEP);
         if (curlen >= dnlen) {
            return INSTALL_DN_ERR; /* overflow */
         }
         snprintf(dnp, dnlen, "%s%s", *s, LDAP_PATHSEP);
         if (!memchr(dnp, 0, dnlen)) return INSTALL_DN_ERR; /*overflow*/
         dnp += curlen;
         dnlen -= curlen;

         s++;
      }

      if (*s == NULL)
      {
         dn[strlen(dn)-strlen(LDAP_PATHSEP)] = 0;
      }

      if (strcasecmp(dn, DEFAULT_ROOT_DN) == 0)
      {
         break;
      }

      err = entry.retrieve(OBJECT_CLASS_FILTER, LDAP_SCOPE_ONELEVEL, dn);

      if (err == NOT_FOUND)
      {
         entry.drop(dn);
         ++i;
      }
      else
      {
         if (i == 0)
         {
            entry.dropAll(dn);
            ++i;
         }
         else
         {
            break;
         }
         testReplicationDeleteStatus(entry.ldap(), dn);
      }
   }

   ldap_value_free(explodedDN);

   return OKAY;
}

NSAPI_EXTERN LdapErrorCode
setupCreateACIForConfigEntry (LdapEntry *configEntry, const char *sieDN)
{
   LdapError err = OKAY;
   char aci[BIG_BUF];

   if (configEntry == NULL || sieDN == NULL || configEntry->ldap() == NULL)
   {
      err = INVALID_INPUT;
   }
   else
   {
      snprintf(aci, sizeof(aci), DEFAULT_SIE_CONFIG_ACI, sieDN);
      if (!memchr(aci, 0, sizeof(aci))) return INSTALL_DN_ERR; /*overflow*/
      configEntry->addAttribute("aci", aci);
   }
   return err;
}

static LdapErrorCode
setupInsertLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN, Bool replaceEntries, Bool entryGranularity)
{
   LdapError ldapError=OKAY;
   char **vals = NULL;
   char dn[MED_BUF];
   NSString tmp;

   LdifEntry ldif(ldifFile);

   if (ldap == NULL || !ldif.isValid())
   {
      return INVALID_INPUT;
   }

   LdapEntry ldapEntry(ldap);

   while (ldif.nextEntry() != -1)
   {
      ldapEntry.clear();

      vals = ldif.getListItems("dn");

      if (vals == NULL && subTreeDN == NULL)
      {
         continue;
      }
      else if (vals == NULL && subTreeDN)
      {
         snprintf(dn, sizeof(dn), "%s", subTreeDN);
      }
      else if (vals && subTreeDN == NULL)
      {
         snprintf(dn, sizeof(dn), "%s", vals[0]);
      }
      else
      {
         snprintf(dn, sizeof(dn), "%s, %s", vals[0], subTreeDN);
      }

      if (!memchr(dn, 0, sizeof(dn))) return INSTALL_DN_ERR; /*overflow*/

      NVPair::freeListItems(vals);

      for (int i = 0; i < ldif.numList(); i++)
      {
         const char *name = ldif.list(i);
         if (!strcasecmp(name, "dn")) continue;
         vals = ldif.getListItems(name);
         if (vals != NULL)
         {
            if (adminSieDN && !strcasecmp(name, "nsClassname"))
            {
               char **s = vals;
               while (*s != NULL)
               {
                  tmp = NSString(*s) + "@" + adminSieDN;
                  ldapEntry.addAttribute(name, tmp);
                  ++s;
               }
            }
            else
            {
               ldapEntry.addAttributes(name, (const char **) vals);
            }
            NVPair::freeListItems(vals);
         }
      }

      if (ldapEntry.exists(dn) == False)
      {
         ldapError = ldapEntry.insert(dn);
      }
      else
      {
         if (replaceEntries)
         {
            ldapError = ldapEntry.replace(dn);
         }
         else if (!entryGranularity)
         {
            ldapError = ldapEntry.update(dn);
         }
      }

      if (ldapError != OKAY)
      {
         setupLogMessage(SETUP_LOG_FATAL, "ldif", "updating dn: %s", dn);
      }
   }

   return ldapError;
}

NSAPI_EXTERN LdapErrorCode
setupReplaceLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN)
{
   return setupInsertLdifEntries(ldap, subTreeDN, ldifFile, adminSieDN, True, False);
}

NSAPI_EXTERN LdapErrorCode
setupUpdateLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN)
{
   return setupInsertLdifEntries(ldap, subTreeDN, ldifFile, adminSieDN, False, False);
}

NSAPI_EXTERN LdapErrorCode
setupAddLdifEntries (const Ldap *ldap, const char *subTreeDN, const char *ldifFile, const char *adminSieDN)
{
   return setupInsertLdifEntries(ldap, subTreeDN, ldifFile, adminSieDN, False, True);
}

NSAPI_EXTERN LdapErrorCode
setupCheckAdminDomain(const Ldap *ldap, const char *adminDomain)
{
   LdapEntry ldapEntry(ldap);
   LdapError err;
   NSString ssContainerDN;
   NSString sieGroup;
   char * domain = NULL;

   // Check for the validity of o=NetscapeRoot
   if (ldapEntry.exists(DEFAULT_ROOT_DN) == False)
   {
      return INVALID_ROOT;
   }

   //
   // if adminDomain is NULL, search for the default ou=NetscapeServers branch where
   // server information are to be stored.
   if (adminDomain == NULL)
   {
      return INVALID_INPUT;
   }
   else
   {
      domain = formAdminDomainDN(adminDomain);
      err = ldapEntry.retrieve(domain);
   }

   if (domain) free(domain);
   return err;
}

NSAPI_EXTERN LdapErrorCode
setupInstallFramework(const Ldap *ldap, const char *adminDomain)
{
   LdapEntry ldapEntry(ldap);
   LdapError err;
   NSString ssContainerDN;
   NSString sieGroup;
   char * domain = NULL;

   // Check for the validity of o=NetscapeRoot
   if (ldapEntry.exists(DEFAULT_ROOT_DN) == False)
   {
      return INVALID_ROOT;
   }

   //
   // if adminDomain is NULL, search for the default ou=NetscapeServers branch where
   // server information are to be stored.
   if (adminDomain == NULL)
   {
      err = ldapEntry.retrieve(DEFAULT_SS_CONTAINER_FILTER, LDAP_SCOPE_SUBTREE, DEFAULT_ROOT_DN);
   }
   else
   {
      domain = formAdminDomainDN(adminDomain);
      err = ldapEntry.retrieve(domain);
   }

   if (err == OKAY)
   {
      ssContainerDN = ldapEntry.entryDN();
   }
   else
   {
      // Cannot find either ou=NetscapeServers or adminDomain
      //
      return INSTALL_DN_ERR;

   }

   // Found ssContainerDN or using a default one
   // Check whether we have write access to this
   err = isSubTreeWritable(ldap, ssContainerDN);

   // Handle the Global Preferences branch
   if (err == OKAY)
   {
      err = setupGlobalPreferences(ldap, ssContainerDN);
   }

   if (domain) free(domain);
   return err;
}

/*********************************************************************
**
** FUNCTION:    isSubTreeWritable
** DESCRIPTION:
**
** INPUTS:      none
** OUTPUTS:     none
** RETURN:
**   True/False
** SIDE EFFECTS:
**      None
** RESTRICTIONS:
**      This Ldap object must have been authenticated
** MEMORY:
**********************************************************************
*/
NSAPI_EXTERN LdapErrorCode
setupIsSubTreeWritable(const Ldap *ldap, const char *subTreeDN)
{
   LdapError err;
   NSString tmp = "cn=Delete Me";
   tmp = tmp + LDAP_PATHSEP + subTreeDN;

   LdapEntry ldapEntry(ldap);

   ldapEntry.addAttribute("objectclass", "top");
   ldapEntry.addAttribute("objectclass", "nsResourceRef");
   ldapEntry.addAttribute("cn", "cn=Delete Me");

   err = ldapEntry.insert(tmp);

   if (err == LDAP_ALREADY_EXISTS || err == OKAY)
   {
      ldapEntry.drop(tmp);
      err = OKAY;
   }

   return err;
}


NSAPI_EXTERN char *
setupGuessAdminDomainName(const char *ldapURL, const char *ldapUser, const char *ldapPwd)
{
   char *domain = NULL;

   if (ldapURL != NULL)
   {
      LdapError err;

      Ldap ldap(err, ldapURL, ldapUser, ldapPwd, NULL, NULL);

      if (err == OKAY)
      {
         LdapEntry domainEntry(&ldap);
         char **explodedDN;

         err = domainEntry.retrieve("(objectclass=nsAdminDomain)", LDAP_SCOPE_ONELEVEL, DEFAULT_ROOT_DN);

         if (err == OKAY)
         {
            explodedDN = ldap_explode_dn((char *) domainEntry.entryDN(), 1);
            domain = strdup(explodedDN[0]);
            ldap_value_free(explodedDN);
         }
      }
   }
   return domain;
}

NSAPI_EXTERN LdapErrorCode
setupAddGlobalUserDirectory (const Ldap *ldap, const char *domain, const char *userDirectoryUrl, const char *bindDN, const char *bindPwd)
{
   char userDirInfoDN[BIG_BUF];
   char url_buffer[BIG_BUF], failover_buffer[BIG_BUF];
   char *temp;

   if (ldap == NULL || domain == NULL || userDirectoryUrl == NULL)
   {
      return INVALID_INPUT;
   }

   // Check to see if domain entered is a DN or just a name
   if (strpbrk(domain, "=,") != NULL)
   {
      snprintf(userDirInfoDN, sizeof(userDirInfoDN), "%s, %s, %s",
               DEFAULT_USER_DIRINFO_RDN,
               DEFAULT_GLOBAL_PREFS_RDN,
               domain);
   }
   else
   {
      snprintf(userDirInfoDN, sizeof(userDirInfoDN), "%s, %s, ou=%s, %s",
               DEFAULT_USER_DIRINFO_RDN,
               DEFAULT_GLOBAL_PREFS_RDN,
               domain,
               DEFAULT_ROOT_DN);
   }
   if (!memchr(userDirInfoDN, 0, sizeof(userDirInfoDN))) return INSTALL_DN_ERR; /*overflow*/

   LdapEntry ldapEntry(ldap);

   if (ldapEntry.exists(userDirInfoDN) == False)
   {
      ldapEntry.setAttribute("objectclass", "top");
      ldapEntry.addAttribute("objectclass", DEFAULT_USER_DIRINFO_OBJECT);
      ldapEntry.setAttribute("cn", DEFAULT_USER_DIRINFO);

      if (sizeof(url_buffer) < strlen(userDirectoryUrl))
      {
         return INSTALL_DN_ERR; /* overflow */
      }
      strncpy(url_buffer, userDirectoryUrl, sizeof(url_buffer));
      strncpy(failover_buffer, "", sizeof(failover_buffer));

      ldapEntry.setAttribute("nsDirectoryURL", url_buffer);
      ldapEntry.setAttribute("nsDirectoryFailoverList", failover_buffer);
      if (bindDN != NULL)
      {
         ldapEntry.setAttribute("nsBindDN", bindDN);
         ldapEntry.setAttribute("nsBindPassword", bindPwd);
      }
      return ldapEntry.insert(userDirInfoDN);
   }
   else
   {
      return OKAY;
   }
}

NSAPI_EXTERN LdapErrorCode
setupGetGlobalUserDirectory(const Ldap *ldap, const char *domain,
   char **url, char **bindDN, char **bindPwd)
{
   char userDirInfoDN[BIG_BUF];
   LdapError err;

   if (ldap == NULL || domain == NULL || url == NULL)
   {
      return INVALID_INPUT;
   }

   *url = NULL;

   if (bindDN) *bindDN = NULL;
   if (bindPwd) *bindPwd = NULL;

   // Check to see if domain entered is a DN or just a name
   if (strpbrk(domain, "=,") != NULL)
   {
      snprintf(userDirInfoDN, sizeof(userDirInfoDN), "%s, %s, %s",
               DEFAULT_USER_DIRINFO_RDN,
               DEFAULT_GLOBAL_PREFS_RDN,
               domain);
   }
   else
   {
      snprintf(userDirInfoDN, sizeof(userDirInfoDN), "%s, %s, ou=%s, %s",
               DEFAULT_USER_DIRINFO_RDN,
               DEFAULT_GLOBAL_PREFS_RDN,
               domain,
               DEFAULT_ROOT_DN);
   }
   if (!memchr(userDirInfoDN, 0, sizeof(userDirInfoDN))) return INSTALL_DN_ERR; /*overflow*/

   LdapEntry ldapEntry(ldap);

   err = ldapEntry.retrieve(userDirInfoDN);

   if (err == OKAY)
   {
      const char *s;

      if (s = ldapEntry.getAttribute("nsDirectoryURL")) {
         char *s2, *temp;
         char buffer[BIG_BUF];
         int slen = strlen(s);

         if (slen >= sizeof(buffer)) {
            return INVALID_URL; /* overflow */
         }
         strncpy(buffer, s, sizeof(buffer));
         
         s2 = ldapEntry.getAttribute("nsDirectoryFailoverList");
         if (s2 && strcmp(s2, ""))
         {
            char *rest;
            /* append to url */
            temp = strrchr(buffer, '/');
            if (NULL != temp) {
               rest = (char *)strrchr(s, '/');
            } else {
               temp = buffer + slen;
               rest = (char *)"";
            }

            snprintf(temp, sizeof(buffer) - slen, " %s%s", s2, rest);
            if (!memchr(temp, 0, sizeof(buffer) - slen)) return INSTALL_DN_ERR; /*overflow*/
         }
         *url = strdup(buffer);
         if (NULL == *url) {
            return INSTALL_DN_ERR;
         }
      }

      if ((s = ldapEntry.getAttribute("nsBindDN")) && bindDN)
         *bindDN = strdup(s);

      if ((s = ldapEntry.getAttribute("nsBindPassword")) && bindPwd)
         *bindPwd = strdup(s);
   }
   return err;
}


NSAPI_EXTERN LdapErrorCode
setupGetUserDirectory(const Ldap *ldap, const char *sieDN,
   char **url, char **bindDN, char **bindPwd)
{
   LdapError err;
   const char *s;

   if (ldap == NULL || sieDN == NULL || url == NULL)
   {
      return INVALID_INPUT;
   }

   *url = NULL;
   if (bindDN) *bindDN= NULL;
   if (bindPwd) *bindPwd = NULL;

   LdapEntry ldapEntry(ldap);
   NSString userDirEntryDN = NSString("cn=configuration") +
                    LDAP_PATHSEP + sieDN;

   err = walkUserDirChain(ldapEntry, userDirEntryDN);

   if (err == OKAY)
   {
      if(s = ldapEntry.getAttribute("nsDirectoryURL")) {
         char *s2, *temp;
         char buffer[BIG_BUF];
         int slen = strlen(s);

         if (slen >= sizeof(buffer)) {
            return INVALID_URL; /* overflow */
         }
         strncpy(buffer, s, sizeof(buffer));

         s2 = ldapEntry.getAttribute("nsDirectoryFailoverList");
         if (s2 && strcmp(s2, ""))
         {
            char *rest;
            /* append to url */
            temp = strrchr(buffer, '/');
            if (NULL != temp) {
               rest = (char *)strrchr(s, '/');
            } else {
               temp = buffer + slen;
               rest = (char *)"";
            }

            snprintf(temp, sizeof(buffer) - slen, " %s%s", s2, rest);
            if (!memchr(temp, 0, sizeof(buffer) - slen)) return INSTALL_DN_ERR; /*overflow*/
         }
         *url = strdup(buffer);
         if (NULL == *url) {
            return INSTALL_DN_ERR; /* overflow */
         }
      }
      s = ldapEntry.getAttribute("nsBindDN");
      if (s && bindDN) *bindDN = strdup(s);
      s = ldapEntry.getAttribute("nsBindPassword");
      if (s && bindPwd) *bindPwd = strdup(s);
   }
   return err;
}

NSAPI_EXTERN LdapErrorCode
setupGetLocalUserDirectory(const Ldap *ldap, const char *domain, const char *hostName, const char * serverRoot, char **url, char **bindDN, char **bindPwd)
{
   NSString targetDN;
   char *criteria;
   LdapErrorCode err = NOT_FOUND;

   if (ldap == NULL || hostName == NULL || serverRoot == NULL ||
        url == NULL)
   {
      return INVALID_INPUT;
   }

   *url = NULL;
   if (bindDN) *bindDN = NULL;
   if (bindPwd) *bindPwd = NULL;

   // Check to see if domain entered is a DN or just a name
   if (strpbrk(domain, "=,") != NULL)
   {
      targetDN = domain;
   }
   else
   {
      targetDN = NSString("ou=") + domain + LDAP_PATHSEP + DEFAULT_ROOT_DN;
   }

   targetDN = NSString("cn=") + hostName + LDAP_PATHSEP + targetDN;

   criteria = formAdminGroupSearchCriteria(serverRoot);

   LdapEntry ldapEntry(ldap);

   err = ldapEntry.retrieve(criteria, LDAP_SCOPE_SUBTREE, targetDN);

   free(criteria);

   if (err == OKAY)
   {
      targetDN = ldapEntry.entryDN();

      err = walkUserDirChain(ldapEntry, targetDN);
   }

   if (err == OKAY)
   {
      const char *s;

      if (s = ldapEntry.getAttribute("nsDirectoryURL")) {
         char *s2, *temp;
         char buffer[BIG_BUF];
         int slen = strlen(s);

         if (slen >= sizeof(buffer)) {
            return INVALID_URL; /* overflow */
         }
         strncpy(buffer, s, sizeof(buffer));

         s2 = ldapEntry.getAttribute("nsDirectoryFailoverList");
         if (s2 && strcmp(s2, ""))
         {
            char *rest;
            /* append to url */
            temp = strrchr(buffer, '/');
            if (NULL != temp) {
               rest = (char *)strrchr(s, '/');
            } else {
               temp = buffer + slen;
               rest = (char *)"";
            }

            snprintf(temp, sizeof(buffer) - slen, " %s%s", s2, rest);
            if (!memchr(temp, 0, sizeof(buffer) - slen)) return INSTALL_DN_ERR; /*overflow*/
         }
         *url = strdup(buffer);
         if (NULL == *url) {
            return INSTALL_DN_ERR; /* overflow */
         }
      }

      s = ldapEntry.getAttribute("nsBindDN");
      if (s && bindDN) *bindDN = strdup(s);
      s = ldapEntry.getAttribute("nsBindPassword");
      if (s && bindPwd) *bindPwd = strdup(s);
   }

   return err;
}

NSAPI_EXTERN LdapErrorCode
setupLdapModify(Ldap *ldap, const char *dn, LDAPMod **mods)
{
  return ldap->modify(dn, mods);
}
