/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version
 * 2.1 of the License.
 *  
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *  
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * END COPYRIGHT BLOCK **/
/***********************************************************************
**
** NAME
**   localToUTF8.cc
**
** DESCRIPTION
**   Converters between the local encoding and UTF-8
**
** NOTES:
**   Public routines:
**     char * localToUTF8(const char *src)
**       -- converts local encoding string src into UTF-8 string utf8.
**          string utf8 is allocated internally, which needs to be freed
**          by the caller.
**
**     char * UTF8ToLocal(const char *utf8)
**       -- converts UTF-8 string utf8 into local encoding string target.
**          string target is allocated internally, which needs to be freed
**          by the caller.
**
**  HISTORY
**  $Log: utf8.cpp,v $
**  Revision 1.1.1.1  2005/07/29 22:16:30  foxworth
**  Importing new setup sdk for open source project
**
**  Revision 1.3  2005/07/29 19:43:27  nkinder
**  164436 - Updated copyrights, licensed code under LGPL, and licensed docs under OPL
**
**  Revision 1.2  2005/03/01 01:45:13  nhosoi
**  Brand the Setup SDK as "Fedora"; updated copyrights
**
**  Revision 1.1.1.1  2005/01/25 18:34:45  cvsadm
**  Move SetupSDK from 6.2 development branch to TRUNK
**
**  Revision 1.1.1.3.4.7  2004/06/01 18:47:11  dboreham
**  Added const cast to make the gcc3.x compiler happy.
**
**  Revision 1.1.1.3.4.6  2003/08/07 20:42:41  norikoyasuo
**  minor indent fix...
**
**  Revision 1.1.1.3.4.5  2003/08/07 20:38:37  norikoyasuo
**  skip iconv if the local character set is UTF-8.
**
**  Revision 1.1.1.3.4.4  2002/12/13 19:34:24  djani
**  Bug 613981: In localToUTF8(), do utf8=strdup(src) only if src is not NULL
**
**  Revision 1.1.1.3.4.3  2002/12/10 23:38:36  djani
**  Bug 613981: get rid of the extra line that got added mistakenly
**
**  Revision 1.1.1.3.4.2  2002/12/10 23:24:27  djani
**  Bug 613981: in case of an error, return the original string. This is part of the libNLS elimination from setupsdk.
**
**  Revision 1.1.1.3.4.1  2002/12/02 20:11:35  djani
**  Elimination of libNLS from setupsdk
**
**  Revision 1.1.1.3  2001/11/06 01:51:32  ahakim
**  XXX AOL/NSCP Copyright XXX
**
**  Revision 1.1.1.2  2001/07/03 18:45:08  svbld
**  (mhein) Changes for x86 port
**
**  Revision 1.1.1.1  1999/03/04 23:28:31  svbld
**  importing setupsdk41 closer to the trunk
**
**  Revision 1.1.4.17.4.5  1998/11/16 23:58:46  cxue
**  compilation error
**
**  Revision 1.1.4.17.4.4  1998/11/14 01:18:11  gbeasley
**  unified api
**
**  Revision 1.1.4.17.4.3  1998/11/11 20:21:22  gbeasley
**  isAsciibuffer
**
**  Revision 1.1.4.20  1998/11/03 23:33:54  cxue
**  check 8bit buffer - isAsciiBuffer
**
**  Revision 1.1.4.19  1998/10/20 20:10:06  cxue
**  bug 330123, setup crahes on unknown locale
**
**  Revision 1.1.4.18  1998/10/06 18:30:04  cxue
**  purify doesn't like free() memory from libnls
**
**  Revision 1.1.4.17  1998/08/20 20:02:07  gbeasley
**  admin domain change
**
**  Revision 1.1.4.16  1998/07/17 02:24:12  pvo
**  Eliminate some memory leaks
**
**  Revision 1.1.4.15  1998/06/05 16:30:27  gbeasley
**  build break
**
**  Revision 1.1.4.14  1998/06/05 03:38:06  pvo
**  118281 - additional code for UTF8 conversion
**
**  Revision 1.1.4.13  1998/04/24 17:23:09  gbeasley
**  unix and nt use nsSetupFree
**
**  Revision 1.1.2.13  1998/04/23 01:04:40  pvo
**  Make NssetupFree shared between Unix and NT
**
**  Revision 1.1.2.12  1998/04/20 23:26:26  gbeasley
**  fix memory leak
**
**  Revision 1.1.4.11  1998/04/20 20:51:33  pvo
**  112701 - problem reconfiguring SIE
**
**/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "utf8.h"

#ifdef XP_WIN32
#define  VC_EXTRALEAN         //  Exclude rarely-used stuff from afxwin.h
#include <afxwin.h>      // MFC core and standard components`
#include <winnls.h>
#endif

#ifdef _WIN32
static char *win_char_converter(const char *instr, int bFromUTF8);
#else
#include <locale.h>
#endif
 
#ifndef _WIN32
#include <iconv.h>
#include <langinfo.h>   /* for nl_langinfo() */
#endif

#ifdef __cplusplus
extern "C" {
#endif

#if defined(_HPUX_SOURCE)
#define LDAPTOOL_CHARSET_UTF8           "utf8"          /* HP/UX */
#else
#define LDAPTOOL_CHARSET_UTF8           "UTF-8"         /* all others */
#endif

#if defined(_HPUX_SOURCE)
#define LDAPTOOL_CHARSET_DEFAULT        "roma8"         /* HP/UX */
#elif defined(__GLIBC__)
#define LDAPTOOL_CHARSET_DEFAULT        "US-ASCII"      /* glibc, e.g. Linux */
#else
#define LDAPTOOL_CHARSET_DEFAULT        "646"           /* all others */
#endif

static char *convertor( const char *src, int fromUtf8 );
#ifndef _WIN32
static const char *GetCurrentCharset(void);
#endif

static int hasHighBit(const char *);

#define BSZ    256

#define DEFAULT_CHARSET  "ISO_8859-1:1987"
char *ldaptool_charset = NULL;

NSAPI_EXTERN char *
localToUTF8(const char *src)
{
    char *utf8;
    if ( src == NULL ) {                /* trivial case # 1 */
        utf8 = NULL;
    } else if ( *src == '\0' ) {        /* trivial case # 2 */
        utf8 = strdup( "" );
    } else {
        utf8 = convertor( src, FALSE );  /* the real deal */
        if ( utf8 == NULL ) {
            utf8 = strdup ( src );
        }
    }

    return utf8;
}


NSAPI_EXTERN char *
UTF8ToLocal(const char *utf8)
{
    char *target = NULL;
    if (utf8 == NULL) {
        target = NULL;
    }
    else if (*utf8 == '\0') {
        target=strdup(utf8);
    }
    else if (hasHighBit(utf8)) {
        target = convertor( utf8, TRUE ); 
        if (target == NULL) {
            target = strdup(utf8);
        }
    }
    else {
        target = strdup( utf8 );
    }

    return target;
}

static int
hasHighBit(const char *s)
{
    unsigned char *p;

    for (p = (unsigned char *)s; *p; p++) {
    if (*p & 0x80)
        return 1;
    }
    return 0;
}


NSAPI_EXTERN int isAsciiBuffer(const char *buf)
{
   int i;
   unsigned char ch;

   if (buf == NULL)
      return True;

   for (i = 0; i < strlen(buf); i ++)
   {
      ch = buf[i];
      if (ch > 0x7F || isascii(ch) == 0) 
         return False;
   }

   return True;
}

#ifdef _WIN32
static char *
convertor( const char *src, int fromUtf8 )
{
    return win_char_converter( src, fromUtf8 );
}
#else /* _WIN32 */

static char *
convertor( const char *src, int fromUtf8 )
{
    static const char   *src_charset = NULL;
    iconv_t             convdesc;
    char                *outbuf, *curoutbuf;
    size_t              inbytesleft, outbytesleft;

    /* Determine the source charset if not already done */
    if ( NULL == src_charset ) {
        char    *locale, *p;

        if ( NULL != ldaptool_charset && 0 != strcmp( ldaptool_charset, "" )) {
            src_charset = ldaptool_charset;
        } else {
            src_charset = GetCurrentCharset();
        }
    }

    if (0 == strcasecmp(src_charset, LDAPTOOL_CHARSET_UTF8))
    {
        outbuf = strdup(src);
        if ( NULL ==  outbuf )
        {
            perror( "convert_to_utf8 - strdup" );
        }
        return outbuf;
    }

    /* Get a converter */
    if (fromUtf8) {
        convdesc = iconv_open( src_charset, LDAPTOOL_CHARSET_UTF8 );
    } else {
        convdesc = iconv_open( LDAPTOOL_CHARSET_UTF8, src_charset );
    }
    if ( (iconv_t)-1 == convdesc ) {
        if ( errno == EINVAL ) {
            fprintf( stderr, "%s: conversion from %s to %s is not supported\n",
                "convertor", src_charset, LDAPTOOL_CHARSET_UTF8 );
        } else {
            perror( src_charset );
        }
        return NULL;
    }

    /* Allocate room for the UTF-8 equivalent (maximum expansion = 6 times) */
/* XXX is that correct? */
    inbytesleft = strlen( src );
    outbytesleft = 6 * inbytesleft + 1;
    if ( NULL == ( outbuf = (char *)malloc( outbytesleft ))) {
        perror( "convert_to_utf8 - malloc" );
        iconv_close( convdesc );
        return NULL;
    }

    curoutbuf = outbuf;
    /*
     * Three steps for a good conversion:
     * 1) Insert the initial shift sequence if any.
     * 2) Convert our characters.
     * 3) Insert the closing shift sequence, if any.
     */
    if ( (size_t)-1 == iconv( convdesc, NULL, NULL,
        &curoutbuf, &outbytesleft ) /* initial shift seq. */
#ifdef LINUX
        || (size_t)-1 == iconv( convdesc, const_cast <char **>(&src), &inbytesleft,
        &curoutbuf, &outbytesleft )     /* convert our chars. */
#else
        || (size_t)-1 == iconv( convdesc, &src, &inbytesleft,
        &curoutbuf, &outbytesleft )     /* convert our chars. */
#endif
        || (size_t)-1 == iconv( convdesc, NULL, NULL,
        &curoutbuf, &outbytesleft )) { /* closing shift seq. */
            perror( "convert_to_utf8 - iconv" );
            iconv_close( convdesc );
            return NULL;
    }
    *curoutbuf = '\0';  /* zero-terminate the resulting string */

    return outbuf;
}

/* returns a malloc'd string */
static const char *
GetCurrentCharset(void)
{
    static char *locale = NULL;
    const char  *charset;

    if ( NULL == locale ) {
        locale = setlocale(LC_CTYPE, "");       /* need to call this once */
    }

    charset = nl_langinfo( CODESET );
    if ( NULL == charset || '\0' == *charset ) {
        charset = LDAPTOOL_CHARSET_DEFAULT;
    }
    return strdup( charset );
}
#endif /* else _WIN32 */

#ifdef __cplusplus
}
#endif  /* __cplusplus */


#ifdef _WIN32
/* returns a malloc'd string */
static char *
win_char_converter(const char *instr, int bFromUTF8)
{
    char  *outstr = NULL;
    int    inlen, wclen, outlen;
    LPWSTR wcstr;
         
    if (instr == NULL)
            return NULL;

    if ((inlen = strlen(instr)) <= 0)
            return NULL;
 
    /* output never becomes longer than input,   XXXmcs: really true?
    ** thus we don't have to ask for the length
    */
    wcstr = (LPWSTR) malloc( sizeof( WCHAR ) * (inlen+1) );
    if (!wcstr)
        return NULL;
 
    wclen = MultiByteToWideChar(bFromUTF8 ? CP_UTF8 : CP_ACP, 0, instr,
                                 inlen, wcstr, inlen);
    outlen = WideCharToMultiByte(bFromUTF8 ? CP_ACP : CP_UTF8, 0, wcstr,
                                  wclen, NULL, 0, NULL, NULL);
 
    if (outlen > 0) {
        outstr = (char *) malloc(outlen + 2);
        outlen = WideCharToMultiByte(bFromUTF8 ? CP_ACP : CP_UTF8, 0, wcstr,
                                      wclen, outstr, outlen, NULL, NULL);
        if (outlen > 0)
            *(outstr+outlen) = _T('\0');
        else
            return NULL;
    }
    free( wcstr );
    return outstr;
}
#endif /* _WIN32 */
